/*
 * @(#)ClientAgent.java
 * KeyDistribution package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */
 
package ksu.cis.example.KeyDistribution;

import java.io.*;
import java.net.*;
import java.security.*;
import javax.net.ssl.*;
import ksu.cis.mom.*;
import java.util.*;

/**
 * <p>ClientAgent class. This agent uses three types of communications.
 * <br>1. Broadcast to register to DistributorAgent.
 * <br>2. SecureUnicastMessage to receive secret key for secured multicast
 * conversation from DistributorAgent.
 * <br>3. SecureMulticastConversation to make secured multicast conversation
 * within the multicast group.
 * @author Chairoj Mekprasertvit
 * @version 1.0, 01/20/2004
 */
public class ClientAgent extends Agent
{
  public static final int unicastPort = 3000;
  public static final int secureUnicastPort = 3001;
  public static final int multicastPort[] = {3002, 3003};
  public static final int broadcastPort = 3004;
  public static final int secureMulticastPort[] = {3005, 3006};
  private ClientAgentGUI gui;
  private Key key;
  private String algorithm;
  private SecureMulticastHandler smh;
  
  public ClientAgent(String name, String broadcastAddress, ClientAgentGUI gui)
  {
    super(
    name, 
    unicastPort, 
    multicastPort, 
    broadcastPort, 
    secureUnicastPort, 
    secureMulticastPort);
    this.gui = gui;
    
    try
    {
      super.group = new InetAddress[5];
      super.group[0] = InetAddress.getByName("239.255.10.11");
      super.group[1] = InetAddress.getByName("239.255.10.12");
      super.broadcast_address = InetAddress.getByName(broadcastAddress);
    }
    catch (Throwable e)
    {
      write("Invalid Internet Address");
    }
    
    BroadcastHandler bh =
      new BroadcastHandler(this, broadcastPort, broadcast_address);
    bh.start();
    
    SecureUnicastHandler suh =
      new SecureUnicastHandler(secureUnicastPort, this);
    suh.start();
  }
  

  /**
   * 
   */
  public void receiveMessage(
    Socket server,
    ObjectInputStream input,
    ObjectOutputStream output)
  {
    

  }
  
  /**<p> Start ClientReturnRegister conversation when receive request to
   * register from the distributor agent.
   * @param bSocket - datagram socket for handling broadcast conversation
   * @param m - message from other agents
   * @param broadcast_queue - message queue for broadcast conversation
   */
  public void receiveBroadcastConversation(
    DatagramSocket bSocket,
    Message m,
    Vector broadcast_queue)
    {
      Thread t;
      if (m.performative.equals("register"))
      {
        t = new Thread(new ClientReturnRegister(bSocket, broadcast_queue, this, broadcastPort, m, gui));
        t.start();
      }
    }

  /**
   * <p>Receive message method for the multicast conversation.
   * Agent must override this method and defines all possible of multicast 
   * conversations here.<br>
   * This is the method that will be called by the <code>MulticastHandler
   * </code> when the agent has received connection starting a new 
   * multicast conversation.</p>
   * @param mSocket - multicast socket for handling multicast conversation
   * @param m - Message from other agents
   * @param multicast_queue - message queue for multicast conversation
   */
  public void receiveMulticastConversation(
    MulticastSocket mSocket,
    Message m,
    Vector multicast_queue)
    {
      
    }
    
  /**
   * <p>Receive multicast join message method for the multicast conversation.
   * Agent can override this method and performs some tasks. For example,
   * agent can keep track of other agents who join the group after itself.<br>
   * This is the method that will be called by the <code>MulticastHandler
   * </code> when the agent has received join message from other agents.</p>
   */
  public void receiveMulticastJoin(Message m)
  {
      
  }  
  
  /**
   * <p>Receive multicast leave message method for the multicast conversation.
   * Agent can override this method and performs some tasks. For example,
   * agent can keep track of other agents who leave.<br>
   * This is the method that will be called by the <code>MulticastHandler
   * </code> when the agent has received leave message from other agents.</p>
   */
  public void receiveMulticastLeave(Message m)
  {
      
  }
  
  /**
   * <p>Receive message method for the secured multicast conversation.
   * Agent must override this method and defines all possible of secured 
   * multicast conversations here.<br>
   * This is the method that will be called by the <code>SecureMulticastHandler
   * </code> when the agent has received connection starting a new secured
   * multicast conversation.</p>
   * @param mSocket - secured multicast socket for handling secured 
   * multicast conversation
   * @param m - message from other agents
   * @param multicast_queue - message queue for secured multicast conversation
   * @param k - symmetric private key.
   * @param algorithm - symmetric key algorithm name
   */
  public void receiveSecureMulticastConversation(
    MulticastSocket mSocket,
    Message m,
    Vector multicast_queue,
    Key k,
    String algorithm)
    {
      System.out.println("receive multicast " + (String)m.content);
      gui.setDisplay((String)m.content);
    }

  /**
   * <p>Receive message method for the secured unicast conversation.
   * Agent must override this method and defines all possible of secured unicast 
   * conversations here.<br>
   * This is the method that will be called by the <code>SecureUnicastHandler
   * </code> when the agent has received connection starting a new 
   * secured unicast conversation.</p>
   * @param server - Secured socket for the secured unicast conversation
   * @param input - ObjectInputStream for the secured unicast conversation
   * @param output - ObjectOutputStream for the secured unicast conversation
   */
  public void receiveSecureUnicastConversation(
    SSLSocket server,
    ObjectInputStream input,
    ObjectOutputStream output)
  {
    Message m;
    Thread t;
    try
    {
      m = (Message) input.readObject();
    
      if (m.performative.equals("GetKey"))
      {
        this.key = (Key) m.content;
        this.algorithm = key.getAlgorithm();
        gui.setDisplay("** Agent has received the secrect key.");
        gui.setDisplay("** Key using: " + key.getAlgorithm());
        
        t = new Thread(new ClientGetKey(server, input, output, this, m));
        t.start();
        
        smh = new SecureMulticastHandler(
          this, secureMulticastPort[1], 1, group[1], key, "DES");
        smh.start();
      }
    }
    catch (ClassNotFoundException cnfex)
    {
      write("** Error class not found in " + this.name);
    }
    catch (IOException ex)
    {
      write("** Error read unicast input in "+ this.name);
    }
  }
  
  public void sendSecuredMulticast(String m)
  {
    Thread init;
    init = new Thread(new SecureMulticastSend(
      this, secureMulticastPort[1], group[1], m, gui, key, algorithm));
    init.start();
  }
    
  /**
   * 
   * @see java.lang.Runnable#run()
   */
  public void run()
  {
    boolean alive = true;
    gui.setDisplay("** Agent Run.");
    while(alive)
    {
    }
  }
}
