/*
 * @(#)Agent.java
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */

package ksu.cis.mom;

import java.net.*;
import javax.net.ssl.*;
import java.security.*;
import java.io.*;
import java.util.*;

/**
 * <p>The Agent class is an abstract class that defines the minimum requirements 
 * for an agent to use agentMom package. This class inherits from 
 * <code>MomObject</code> class. It must be runnable as a separate thread, 
 * which requires a run method. It also has two required parameters that must 
 * be set for each agent, the name of the agent and the port number on which 
 * its message handler will listen for incoming messages.
 * <p>Also, this class is associated with the
 * <code>MessageHandler</code>, <code>MulticastHandler</code>, 
 * <code>BroadcastHandler</code>, <code>SecureUnicastHandler</code> and
 * <code>SecureMulticastHandler</code> class. 
 * <p>An agent needs to start the correct handler to be able to receive the 
 * start of conversation from other agents. Furthermore, each agent must 
 * implement/override receive message method for each type of conversation. 
 * For unicast conversation, the agent must implement the receiveMessage method. 
 * For the other conversation, the agent must override the receive method 
 * corresponding to each conversation. The receive method will be called by the 
 * corresponding message handler when the agent has received a connection 
 * starting a new conversation. For example, the corresponding handler of 
 * receiveMulticastConversation method is MulticastHandler class, and the 
 * corresponding handler of receiveMessage method MessageHandler class.
 * 
 * <p>History:<br>
 * 4 October 1999 - Creation Date<br>
 * 3 May 2001 - extended to allow conversations to work with either agents or 
 * conversations as their parents<br>
 * Summer 2003 - Modify and Add more comments suitable for JavaDoc.<br>
 * Fall 2003 - Add receiveMulticast, receiveBroadcast, receiveSecureUnicast
 * receiveSecureMulticast method.</p>
 * 
 *
 * @author Dr. Scott DeLoach and Chairoj mekprasertvit
 * @version 2.0, 08/23/2003
 * @see ksu.cis.mom.MomObject
 * @see ksu.cis.mom.MessageHandler
 * @see ksu.cis.mom.MulticastHandler
 * @see ksu.cis.mom.BroadcastHandler
 * @see ksu.cis.mom.SecureUnicastHandler
 * @see ksu.cis.mom.SecureMulticastHandler
 * @since agentMom 1.0
 */

public abstract class Agent extends MomObject implements Runnable, Serializable
{
  /**
   * Message queue for broadcast conversation.
   */
  public Vector broadcast_queue;
  
  /**
   * Vector of internal components that are active.
   */
  private Vector components = new Vector();
  
  /**
   * Message queue for multicast conversation.
   */
  public Vector multicast_queue;
  
  /**
   * Message queue for secured multicast conversation.
   */
  public Vector secure_multicast_queue;

  /**
   * Default constructor. This is an empty constructor and does not require any 
   * argument.
   */
  protected Agent()
  {
  }

  /**
   * Constructor for using only unicast conversation. It takes two argument, 
   * String of agent name and integer of unicast port.
   * @param name - String name of the agent
   * @param port - Integer port of unicast
   */
  public Agent(String name, int port)
  {
    this.name = name;
    this.port = port;
  }

  /**
   * <p>Constructor for using any or all of five different conversations. It 
   * takes six argument, String of agent name, integer of port used in each 
   * conversation. If any port is assigned to be less than one, then it indicates 
   * the conversation is not going to be used.<br>
   * Note that the arguments multicast_port and secure_multicast_port are an 
   * array of integer type. It is because we allows agent to subscribe to 
   * multiple groups. Thus, one port is used for one group. Assigning the first 
   * element less than one indicates that the multicast will not be used.</p>
   * @param name - String name of the agent
   * @param unicast_port - Integer port number used for unicast conversation
   * @param multicast_port - Array of integer port number used for multicast 
   * conversation 
   * @param broadcast_port - Integer port number used for broadcast conversation
   * @param secure_unicast_port - Integer port number used for secured unicast
   * conversation
   * @param secure_multicast_port - Array of integer port number used for 
   * secured multicast conversation 
   */
  public Agent(
    String name,
    int unicast_port,
    int multicast_port[],
    int broadcast_port,
    int secure_unicast_port,
    int secure_multicast_port[])
  {
    this.name = name;
    if(unicast_port > 0)
    {
      this.port = unicast_port;
    }
    if(multicast_port[0] > 0)
    {
      this.multicast_port = multicast_port;
      this.multicast_queue = new Vector(50);
    }
    if(broadcast_port > 0)
    {
      this.broadcast_port = broadcast_port;
      this.broadcast_queue = new Vector(50);
    }
    if(secure_unicast_port > 0)
    {
      this.secure_unicast_port = secure_unicast_port;
    }
    if(secure_multicast_port[0] > 0)
    {
      this.secure_multicast_port = secure_multicast_port;
      this.secure_multicast_queue = new Vector(50);
    }
    
  }

  /**
   * <p>Receive message method for the unicast conversation.
   * Agent must implement this method and defines all possible of unicast 
   * conversations here.<br>
   * This method will be called by the <code>MessageHandler</code> when the agent
   * has received connection starting a new conversation.</p>
   * @param server - Socket for the unicast conversation
   * @param input - ObjectInputStream for the unicast conversation
   * @param output - ObjectOutputStream for the unicast conversation
   */
  public abstract void receiveMessage(
    Socket server,
    ObjectInputStream input,
    ObjectOutputStream output);

  /**
   * <p>Receive message method for the broadcast conversation. Agent must 
   * override this method and defines all possible of secured multicast 
   * conversations here.<br>
   * This is the method that will be called by the <code>BroadcastHandler
   * </code> when the agent has received connection starting a new broadcast
   * conversation.</p>
   * @param bSocket - datagram socket for handling broadcast conversation
   * @param m - message from other agents
   * @param broadcast_queue - message queue for broadcast conversation
   */
  public void receiveBroadcastConversation(
    DatagramSocket bSocket,
    Message m,
    Vector broadcast_queue)
    {
      
    }

  /**
   * <p>Receive message method for the multicast conversation.
   * Agent must override this method and defines all possible of multicast 
   * conversations here.<br>
   * This is the method that will be called by the <code>MulticastHandler
   * </code> when the agent has received connection starting a new 
   * multicast conversation.</p>
   * @param mSocket - multicast socket for handling multicast conversation
   * @param m - Message from other agents
   * @param multicast_queue - message queue for multicast conversation
   */
  public void receiveMulticastConversation(
    MulticastSocket mSocket,
    Message m,
    Vector multicast_queue)
    {
      
    }
    
  /**
   * <p>Receive multicast join message method for the multicast conversation.
   * Agent can override this method and performs some tasks. For example,
   * agent can keep track of other agents who join the group after itself.<br>
   * This is the method that will be called by the <code>MulticastHandler
   * </code> when the agent has received join message from other agents.</p>
   */
  public void receiveMulticastJoin(Message m)
  {
      
  }  
  
  /**
   * <p>Receive multicast leave message method for the multicast conversation.
   * Agent can override this method and performs some tasks. For example,
   * agent can keep track of other agents who leave.<br>
   * This is the method that will be called by the <code>MulticastHandler
   * </code> when the agent has received leave message from other agents.</p>
   */
  public void receiveMulticastLeave(Message m)
  {
      
  }
  
  /**
   * <p>Receive message method for the secured multicast conversation.
   * Agent must override this method and defines all possible of secured 
   * multicast conversations here.<br>
   * This is the method that will be called by the <code>SecureMulticastHandler
   * </code> when the agent has received connection starting a new secured
   * multicast conversation.</p>
   * @param mSocket - secured multicast socket for handling secured 
   * multicast conversation
   * @param m - message from other agents
   * @param multicast_queue - message queue for secured multicast conversation
   * @param k - symmetric private key.
   * @param algorithm - symmetric key algorithm name
   */
  public void receiveSecureMulticastConversation(
    MulticastSocket mSocket,
    Message m,
    Vector multicast_queue,
    Key k,
    String algorithm)
    {
      
    }

  /**
   * <p>Receive message method for the secured unicast conversation.
   * Agent must override this method and defines all possible of secured unicast 
   * conversations here.<br>
   * This is the method that will be called by the <code>SecureUnicastHandler
   * </code> when the agent has received connection starting a new 
   * secured unicast conversation.</p>
   * @param server - Secured socket for the secured unicast conversation
   * @param input - ObjectInputStream for the secured unicast conversation
   * @param output - ObjectOutputStream for the secured unicast conversation
   */
  public void receiveSecureUnicastConversation(
    SSLSocket server,
    ObjectInputStream input,
    ObjectOutputStream output)
    {
      
    }

  /**
   * <p>An agent must be runnable as a separate thread, and each agent must 
   * implement this method.</p>
   */
  public abstract void run();

  /**
   * <p>Method for sending message between components within an agent. This 
   * method simply broadcast message to all active components of the agent. 
   * @param m - message for internal conversation among the component.
   */
  public void sendInternal(Message m)
  {
    if (components != null)
    {
      for (int i = 0; i < components.size(); i++)
      {
        Component c = (Component) components.elementAt(i);
        c.enqueueInternal(m);
      }
    }
  }

  /**
   * <p>Method for printout a String information on terminal screen.
   * @param s String to print out on the terminal screen.
   */
  public void write(String s)
  {
    System.out.println(s);
  }

}
