/*
 * @(#)Message.java
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */
package ksu.cis.mom;

/**
 * <p><code>Message</code> class definition. <code>Message</code> class defines 
 * the field used in the message passed back and forth between agents. Not all 
 * of the fields need to be used. Note that these fields are derived from the 
 * fields in a KQML message and only some of them are necessary.
 * <p>When a conversation calls the sendMessage method, it automatically fill the 
 * sender, host, and port fields using the parent agents name and port 
 * attributes and automatically gets the host name from the system. 
 * The other fields of interest in an agentMom message are the performative and 
 * content fields. The performative field describe the action that the message 
 * intends and is used in the agent and conversation classes to<br>
 * (1) determine the type of conversation being requested and <br>
 * (2) to control the execution of a conversation in the run method.</p>
 * <p>agentMom does not have any specific performative types. The user define 
 * any performative they feel are necessary. The content of an agentMom message 
 * is also very general. Basically, the message passes any valid Java object 
 * type. This can be as simple as a string, or a more complex object that 
 * encapsulates a number of attribute types. These complex objects can be used 
 * to pass multiple parameters in a single message as shown in the registration 
 * class below.</p>
 * <pre>
 * public class ComplexObject implements Serializable
 * {
 *   String agent;
 *   String host;
 *   int port;
 *   String service;
 *   public ComplexObject(String a, String h, int p, String ser)
 *   {
 *     agent = a;
 *     host = h;
 *     port = p;
 *     service = ser;
 *   }
 * }
 * </pre>
 * <p>This class encapsulates four parameters (three strings and an integer) 
 * that can be assigned to message content field.
 * <p>Note that in order to pass an object across a socket connection, 
 * it must implement the interface Serializable.
 * <p>History:<br>
 * 09/01/2003: Add more comments suitable for JavaDoc.<br>
 * @author Scott DeLoach and Chairoj Mekprasertvit
 * @version 2.0, 08/23/2003
 * @see ksu.cis.mom.Agent
 * @since agentMom 1.0
 */

public class Message implements java.io.Serializable
{
  /**
   * Host name that this message is sent to.
   */
	public String host = null;
  
	/**
   * Port number used for the message.
	 */
  public int port = 0;
  
	/**
   * Name of the sender (agent's name). 
	 */
  public String sender = null;
  
	/**
   * Name of the receiver
	 */
  public String receiver = null;
  
	/**
   * Describe the action that the message intends. The user can define 
   * any performative they feel are necessary.
	 */
  public String performative = null;
  
  /**
   * Specify whether the sender will never deny the meaning of the performative.
   */
	public String force = null;

  /**
   * The expected label in a reply
   */
	public String inreplyto = null; 
  
  /**
   * Name of representation language of the content.
   */
	public String language = null;
  
  /**
   * Name of the ontoloty used in the content
   */
	public String ontology = null; 
  
  /**
   * Whether the sender expects a reply, and if so, a label for the reply.
   */
	public String replywith = null;

  /**
   * Support for complex object that encapsulates a number of attribute types. 
   * These complex objects can be used to pass multiple parameters in a single 
   * message. Note that in order to pass an object across a socket connection, 
   * it must implement the interface Serializable.
   */
	public Object content = null;

  /**
   * Meesage class Constructor.
   */
	public Message()
	{
		super();
	}
  
  /**
   * Return content field in form of the Object.
   * @return the content field in form of the Object.
   */
	public Object getContent()
	{
		return content;
	}
  
  /**
   * Return the performative field in form of the String.
   * @return the performative field in form of the String.
   */
	public String getPerformative()
	{
		return performative;
	}
  
  /**
   * Return the receiver field in form of the String.
   * @return the receiver field in form of the String.
   */
	public String getReceiver()
	{
		return receiver;
	}
  /**
   * Return the sender field in form of the String.
   * @return sender field in form of the String.
   */
	public String getSender()
	{
		return sender;
	}

  /**
   * Set content field of the message in form of the Object.
   * @param cont - content of the message in form of the Object.
   */
	public void setContent(Object cont)
	{
		content = cont;
	}
  
  /**
   * Set content field of the message in form of the String.
   * @param cont - content of the message in form of the String.
   */
	public void setContent(String cont)
	{
		content = cont;
	}
  
  /**
   * Set performative of the message.
   * @param perf - performative of the message in form of the String.
   */
	public void setPerformative(String perf)
	{
		performative = perf;
	}
  
  /**
   * Set the receiver's name.
   * @param name - name of the receiver.
   */
	public void setReceiver(String name)
	{
		receiver = name;
	}
  
  /**
   * Set the sender's name.
   * @param name - name of the sender.
   */
	public void setSender(String name)
	{
		sender = name;
	}
}
