/*
 * @(#)MulticastHandler.java 
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */
 
package ksu.cis.mom;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.SocketTimeoutException;
import java.util.Vector;

/**
 * <p>MulticastHandler is responsible for initializing and starting multicast
 * socket, including joining/leaving multicast group. In generally, it handle 
 * multicast connection with other agents. It uses the MulticastSocket class to 
 * subscribe to multicast group. 
 * <p>When an agent is created, it needs to creates a new multicast handler 
 * thread to be able to receive a start of multicast conversation from other
 * agents.
 * When MulticastHandler class is created, it creates the multicast socket and 
 * join to specified multicast group. Then, it automatically sends a multicast 
 * message to the group indicating that this agent has join the group. When a
 * MulticastHandler receives the message indicating the join, it calls the
 * receiveMulticastJoin method in agent class. 
 * Below is how an agent can start the MessageHandler 
 * class.
 * <pre>
 * MulticastHandler mh = new MessageHandler(this, this.port, 0, multicast_address);
 * mh.start();
 * </pre>
 * <p>To create the MessageHandler, it requires four/five parameters, 
 * port number, a reference to parent agent object, time to live of multicast 
 * packet, multicast address and/or maximum datagram received packet size.
 * <br>Note that the multicast address is actually a class D IP addresses that
 * are in the range 224.0.0.0 to 239.255.255.255.<br>
 * When this class is started, the multicast handler starts a multicast socket 
 * on the indicated port and waits for messages from other agents.
 * When multicast handler receives a message, it will check whether the message
 * is a start of new conversation/join/leave/conversation message. If it is a 
 * start of conversation, the MulticastHandler simply calls the parent agent's 
 * receiveMulticastcastConversation method with the multicast socket, received 
 * message and multicast message queue.
 * The agent then verify the received message and start an appropriate 
 * conversation. If the received message is for any multicast conversation 
 * class, it adds the message to the multicast message queue. Then the 
 * multicast conversation class can get the message from this queue later.
 * <p>Note that the multicast communication has loopback effect(message also 
 * send to itself), but the MulticastHandler will ignore this message.
 * <p>History:<br>
 * 06/12/2003 - creation date<br>
 * @author Scott DeLoach and Chairoj Mekprasertvit
 * @version 1.0, 07/23/2003
 * @since agentMom 2.0
 */

public class MulticastHandler extends Thread
{

  /**
   * Default maximum size of received multicast packet. 
   */
  public int maxSize = 1024;
  
  /**
   * To indicate that the message is a request to start conversation. 
   */
  public static final int START_CONVERSATION = 0;
  
  /**
   * To indicate that the message is not a request to start conversation.
   */
  public static final int CONTENT = 1;
  
  /**
   * Identifies a JOIN multicast PDU
   */
  public static final int JOIN = 2;
  
  /**
   * Identifies a LEAVE multicast PDU
   */
  public static final int LEAVE = 3;
  
  /**
   * Reference to parent agent class.
   */
  public Agent parent;
  
  /**
   * Time-to-live for multicast packets sent out on this MulticastSocket 
   * in order to control the scope of the multicasts.
   */
  public int TimeToLive;

  /**
   * Multicast socket used to send and receive multicast message
   */
  public MulticastSocket mSocket;
  
  /**
   * Multicast port
   */
  public int multicast_port;
  
  /**
   * Message queue for multicast conversation.
   * It is a Vector type that stores multicast message sent to this agent.
   */
  public Vector multicast_queue;
  
  /**
   * multicast group address
   */
  protected InetAddress group;
  
  /**
   * Flag control receive thread execution
   */
  protected boolean leave;
  
  
  /**
   * Default constructor for MulticastHandler. It is responsible for 
   * initializing necessary variables and creates multicast socket.
   * @param p - reference to parent agent class.
   * @param port - port number for listening for multicast message.
   * @param ttl - time to live of datagram packet.
   * @param group - multicast address, in the range 224.0.0.0-239.255.255.255
   */
  public MulticastHandler(Agent p, int port, int ttl, InetAddress group)
  {
    // Initialize variables
    super("MulticastHandler"); //call to superclass(thread class) constructor
    this.parent = p;
    this.multicast_port = port;
    this.multicast_queue = p.multicast_queue;
    this.leave = false;
    this.TimeToLive = ttl;
    this.group = group;

    // Validate port. Port used must be greater than 1024.
    if (multicast_port <= 1024)
      throw new IllegalArgumentException(
        " ** Bad port number given to constructor for agent: " + p.name);

    // Connect MulticastSocket to the given port
    try
    {
      mSocket = new MulticastSocket(multicast_port);
      mSocket.setTimeToLive(TimeToLive);
      mSocket.setLoopbackMode(true);
      mSocket.joinGroup(group);
      sendJoin();
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to multicast port on port"
          + multicast_port); 
      System.exit(1);
    }
    System.out.println(
          "Agent " + p.name + " MulticastHandler Started on port " + port);
  }
  
  /**
   * Default constructor for MulticastHandler. It is responsible for 
   * initializing necessary variables and creates multicast socket. This
   * constructor allows to specify the size of received multicast packet.
   * <p>Note that the received packet size must be greater than or equal to
   * the sent packet. Otherwise, packet will be partially received.
   * @param p reference to parent agent class.
   * @param port - port number for listening for broadcast message.
   * @param ttl - time to live of datagram packet.
   * @param group - multicast address, in the range 224.0.0.0-239.255.255.255
   * @param packetSize - user defined maximum size in byte of received 
   * multicast packet.
   */
  public MulticastHandler(
    Agent p,
    int port,
    int ttl,
    InetAddress group,
    int packetSize)
  {
    // Initialize variables
    super("MulticastHandler"); //call to superclass(thread class) constructor
    this.parent = p;
    this.multicast_port = port;
    this.multicast_queue = p.multicast_queue;
    this.leave = false;
    this.TimeToLive = ttl;
    this.group = group;
    this.maxSize = packetSize;

    // Validate port. Port used must be greater than 1024.
    if (multicast_port <= 1024)
      throw new IllegalArgumentException(
        " ** Bad port number given to constructor for agent: " + p.name);

    // Connect MulticastSocket to the given port
    try
    {
      mSocket = new MulticastSocket(multicast_port);
      mSocket.setTimeToLive(TimeToLive);
      mSocket.setLoopbackMode(true);
      mSocket.joinGroup(group);
      sendJoin();
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to multicast port on port"
          + multicast_port);
      System.exit(1);
    }
    System.out.println(
          "Agent " + p.name + " MulticastHandler Started on port " + port);
  }
  
  /**
   * <p>Sends a multicast JOIN message to notify other agents.
   * We need to convert Message object to Byte type to be able to send it using
   * multicast socket.
   * @throws IOException
   */
  public void sendJoin()throws IOException
  {
    write("Join Group and Send join message.");
    // Initialize output stream
    ByteArrayOutputStream bStream = new ByteArrayOutputStream();
    ObjectOutputStream os = 
      new ObjectOutputStream(new BufferedOutputStream(bStream));
      
    // Initialize Message
    Message m = new Message();
    m.sender = parent.name;
    m.setContent("Join group");
    os.flush();
    os.writeInt(JOIN);
    os.writeObject(m);
    os.flush();

    // Retrives byte array
    byte[] mesg_in_byte = bStream.toByteArray();
    DatagramPacket p = new DatagramPacket(
      mesg_in_byte, 
      mesg_in_byte.length, 
      this.group, 
      this.multicast_port);
    mSocket.send(p);
  }
  
  /**
   * <p>Sends a multicast LEAVE message to notify other agents.
   * We need to convert Message object to Byte type to be able to send it using
   * multicast socket.
   * @throws IOException
   */
  public void sendLeave()throws IOException
  {
    // Initialize output stream
    ByteArrayOutputStream bStream = new ByteArrayOutputStream();
    ObjectOutputStream os = 
      new ObjectOutputStream(new BufferedOutputStream(bStream));
      
    // Initialize Message
    Message m = new Message();
    m.sender = parent.name;
    m.setContent("Leave group");
    os.flush();
    os.writeInt(LEAVE);
    os.writeObject(m);
    os.flush();

    // Retrives byte array
    byte[] mesg_in_byte = bStream.toByteArray();
    DatagramPacket p = new DatagramPacket(
      mesg_in_byte, 
      mesg_in_byte.length, 
      this.group, 
      this.multicast_port);
    mSocket.send(p);
    mSocket.leaveGroup(group);
    mSocket.close();
  }  
  
  /**
   * Run thread method for MulticastHandler.It simply waits for incoming 
   * multicast message from other agents. Then, it checks the message whether 
   * it is a start of new multicast conversation, or a message for multicast 
   * conversation class. 
   * <p>Note that this is a blocking read message.
   */
  public void run()
  {
    System.out.println("Multicast Running on " + multicast_port);
    boolean alive = true;
    while(alive)
    {
      try
      {
        // Local variables
        Message m = new Message();
        byte[] b = new byte[maxSize];
        int header = 0;
        DatagramPacket packet = new DatagramPacket(b, b.length);
        
        packet.setLength(b.length);
        mSocket.setSoTimeout(1000);
        this.checkLeave();
        mSocket.receive(packet);
        ByteArrayInputStream bStream = new ByteArrayInputStream(b);
        ObjectInputStream is = 
          new ObjectInputStream(new BufferedInputStream(bStream));
        header = is.readInt();
        m = (Message) is.readObject();
        
        // ignore message if the sender is itself (prevent loopback effect)
        if (packet.getAddress().getHostAddress().equals(InetAddress.getLocalHost().getHostAddress()))
        {
          write("Loopback - IGNORED");
        }
        else if(header == JOIN)
        {
          write("Receive Join from " + m.sender);
          parent.receiveMulticastJoin(m);
        }
        else if(header == LEAVE)
        {
          parent.receiveMulticastLeave(m);
        }
        else if(header == CONTENT)
        {
          this.queueMessage(m);
        }
        else if(header == START_CONVERSATION)
        {
          parent.receiveMulticastConversation(mSocket, m, multicast_queue);
        }
      } 
      catch (SocketTimeoutException ste)
      {
        //this.write("** Time out check.");
        this.checkLeave();
      }
      catch (IOException e)
      {
        this.write(" ** Multicast Socket is closed on port: " + multicast_port);
        alive = false;
      }
      catch (ClassNotFoundException cnfex)
      {
        this.write(" ** ClassNotFoundException from MulticastHandler.");
      }

    }// end while loop
  }// End of run method
  
  /**
   * Method for easily printout information on screen.
   * @param s java.lang.String
   */
  public void write(String s)
  {
    System.out.println(s);
  }
  
  /**
   * Method for adding new message to the queue.
   * @param m - message to be retrived by multicast conversation.
   */
  public void queueMessage(Message m)
  {
    this.multicast_queue.addElement(m);
  }

  /**
   * Set value for leave.
   * @param leave - if value is true, it will make the handler thread to
   * stop the next time this value is checked in the run method.
   */
  public void setLeave(boolean leave)
  {
    this.leave = leave;
  }
  
  /**
   * To check the value of leave attribute.
   */
  public void checkLeave()
  {
    try
    {
      if (leave)
        this.sendLeave();
    }
    catch (IOException ioex)
    {
      this.write("** Can't leave the group. " + ioex.toString());
    }
  }

}// End of class MulticastHandler
