/*
 * @(#)SecureMulticastConversation.java
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */
 
package ksu.cis.mom;

import java.io.*;
import java.net.*;
import java.util.*;
import java.security.*;
import javax.crypto.*;

/**
 * <p>The SecureMulticastConversation class is an abstract class that actually 
 * carries out the secured multicastcast message to all agents subscribed to 
 * the same group as the sender.
 * <br>This class provides two main services, read multicast message and send 
 * encrypted multicast message. It is responsible for passing the messages 
 * back and forth using multicast socket. 
 * <p>There are really two types of conversation classes that can be derived 
 * from the SecuredMulticastConversation class, one for the conversation 
 * initiator and one for the conversation respondent. The basic difference 
 * lies in which constructor is used and the details in the abstract run 
 * method, which must be implemented in the concrete class derived from the 
 * SecureMulticastConversation class.
 * <p>History<br>
 * 12/01/2003: creation date<br>
 * 11/27/2003: Add more comments suitable for JavaDoc.<br> 
 * @author Scott DeLoach and Chairoj Mekprasertvit
 * @version 1.0 12/27/2003
 * @since agentMom 2.0
 */
public abstract class SecureMulticastConversation extends AgentConversation
{
  /**
   * Default maximum size of received multicast packet. 
   */
  public int maxSize = 1024;

  /**
   * To indicate that the message is a request to start conversation. 
   */
  public static final int START_CONVERSATION = 0;
  
  /**
   * To indicate that the message is not a request to start conversation.
   */
  public static final int CONTENT = 1;
  
  /**
   * Multicast socket used to send and receive multicast prorocol
   */
  public MulticastSocket mSocket;

  /**
   * Time-to-live for multicast packets sent out on this MulticastSocket 
   * in order to control the scope of the multicast.
   */
  protected int TimeToLive;

  /**
   * Message queue for multicast conversation.
   * It is a Vector type that stores multicast message sent to this agent.
   */
  protected Vector secure_multicast_queue;

  /**
   * Name of the conversation. This name must be unique from others multicast
   * conversations within agent since it is used for identifying the 
   * destined conversation when multicast message is received.
   */
  private String conversation_name;

  /**
   * Key class to store the secret key for encryption and decryption.
   */
  protected Key key = null;
  
  /**
   * Key class to store the secret key for encryption and decryption.
   */
  protected String algorithm = null;
  
  /**
   * Cipher class to perform encrption and decryption.
   */
  protected Cipher cipher = null;

  /**
   * Default Conversation constructor.
   * @param c - Reference to parent MomObject class.
   */
  public SecureMulticastConversation(MomObject c)
  {
    super(c);
  }

  /**
   * Constructor for conversation initiator.
   * @param c - reference to parent class.
   * @param group - multicastt address (Class D IP address) .
   * @param port - secured multicast port.
   * @param secure_multicast_queue - message queue for multicast conversation.
   * @param key - symmetric private key
   * @param algorithm - Symmetric key algorithm.
   */
  public SecureMulticastConversation(
    MomObject c,
    InetAddress group,
    int port,
    Vector secure_multicast_queue,
    Key key,
    String algorithm)
  {
    super(c, group.getHostAddress(), port);
    this.parent = c;
    this.secure_multicast_queue = secure_multicast_queue;
    this.key = key;
    this.algorithm = algorithm;
    
    try
    {
      this.mSocket = new MulticastSocket();
      //this.mSocket.setLoopbackMode(true);
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to multicast socket"); 
      System.exit(1);
    }
    
    // Initialize Cipher class
    try
    {
      this.cipher = Cipher.getInstance(this.algorithm);
    }
    catch (NoSuchAlgorithmException nsae)
    {
      System.out.println("** Algorithm " + algorithm + " cannot be found.");
      System.exit(1);
    }
    catch (NoSuchPaddingException nspe)
    {
      System.out.println("Padding scheme in " + algorithm + " is not available.");
      System.exit(1);
    }
  }

  /**
   * Constructor for conversation respondent.
   * @param c - reference to parent class.
   * @param group - multicastt address (Class D IP address).
   * @param port - secured multicast port.
   * @param secure_multicast_queue - message queue for multicast conversation.
   * @param m - ksu.cis.mom.Message
   * @param key - symmetric private key.
   * @param algorithm - symmetric key algorithm.
   */
  public SecureMulticastConversation(
    MomObject c, 
    InetAddress group, 
    int port,
    Vector secure_multicast_queue,
    Message m,
    Key key,
    String algorithm)
  {
    super(c, m);
    this.connectionHost = group.getHostAddress();
    this.connectionPort = port;
    this.secure_multicast_queue = secure_multicast_queue;
    this.key = key;
    this.algorithm = algorithm;
    
    try
    {
      this.mSocket = new MulticastSocket();
      this.mSocket.setLoopbackMode(false);
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to multicast socket"); 
      System.exit(1);
    }
    
    //  Initialize Cipher class
    try
    {
      this.cipher = Cipher.getInstance(this.algorithm);
      this.mSocket = new MulticastSocket();
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to multicast socket"); 
      System.exit(1);
    }
    catch (NoSuchAlgorithmException nsae)
    {
      System.out.println("** Algorithm " + algorithm + " cannot be found.");
      System.exit(1);
    }
    catch (NoSuchPaddingException nspe)
    {
      System.out.println("Padding scheme in " + algorithm + " is not available.");
      System.exit(1);
    }
  }
  
  /**
   * <p>This method is responsible for sending the Message m through the
   * multicast socket. It also automatically fill the sender, host, port, 
   * replywith and inreplyto fields in the Message object m using the parent 
   * agents name and port attributes and automatically gets the host name from 
   * the system. Furthermore, the message will be encrypted using the specified
   * sysmetric key algorithm.
   * <br>We need to convert Message object to Byte type, and then encrypts the byte
   * information to be able to send encrypted message via multicast socket.
   * <br>Note that to send message request a start of new conversation the 
   * correct method is startConversation.
   * @param m - message to send
   * @param replywith - name of the originating conversation.
   * @param inreplyto - name of the destined conversation on the other sides of 
   * agent.
   */
  public void sendMessage(
    Message m, 
    String replywith, 
    String inreplyto) throws IOException
  {
    // Initialize Message
    m.sender = parent.name;
    this.write("Sender is " + parent.getClass().getName());
    
    /* Specify the originating conversation so that the receiving conversation
        knows where to reply to. */
    m.replywith = replywith;
    
    // Specify the destined conversation name.
    m.inreplyto = inreplyto;
    m.port = connectionPort;

    // To convert object to byte array
    ByteArrayOutputStream bStream = new ByteArrayOutputStream();
    ObjectOutputStream os = 
      new ObjectOutputStream(new BufferedOutputStream(bStream));
    
    this.write("Sending message \"" + m.performative + "\" from " + connectionPort);
    if (m.getContent() != null)
      this.write("Content is: " + m.getContent().getClass().getName());
    os.flush();
    os.writeInt(CONTENT);
    os.writeObject(m);
    os.flush();

    // Retrives byte array
    byte[] mesg_in_byte = bStream.toByteArray();
    System.out.println("** Multicast Message before encryption in bit:");
    System.out.println("Bit: ");
    for(int i=0; i<mesg_in_byte.length; i++)
    {
      System.out.print(mesg_in_byte[i]);
    }
    // Encrypt information and send
    try
    {
      cipher.init(Cipher.ENCRYPT_MODE, key);
      byte[] encrypted_mesg_in_byte =  cipher.doFinal(mesg_in_byte);
      DatagramPacket p = new DatagramPacket(
        encrypted_mesg_in_byte, 
        encrypted_mesg_in_byte.length, 
        InetAddress.getByName(connectionHost), 
        connectionPort);   
      mSocket.send(p);
    }
    catch(InvalidKeyException ike)
    {
      System.out.println("** InvalidKeyException in SecureMulticastHandler");
    }
    catch(BadPaddingException bpe)
    {
      System.out.println("** BadPaddingException in SecureMulticastHandler");
    }
    catch(IllegalBlockSizeException ibse)
    {
      System.out.println("** IllegalBlockSizeException in SecureMulticastHandler");
    }
  }
  
  /**
  * <p>This method is responsible for sending message request a start of new 
  * conversation. The Message m is encrypted and then sent through the 
  * multicast socket. 
  * <br>It also automatically fill the sender, host, port, replywith and inreplyto 
  * fields in the Message object m using the parent agents name and port 
  * attributes and automatically gets the host name from the system. 
  * Furthermore, the message will be encrypted using the specified
  * sysmetric key algorithm.
  * <br>We need to convert Message object to Byte type, and then encrypts the byte
  * information to be able to send encrypted message via multicast socket.
  * @param m - message to send
  * @param replywith - name of the originating conversation.
  * @param inreplyto - name of the destined conversation on the other sides of 
  * agent.
  * @throws IOException
  */
  public void startConversation(Message m, String replywith, String inreplyto)
    throws IOException
  {
    // Initialize Message
    m.sender = parent.name;
    this.write("Sender is " + parent.getClass().getName());

    /* Specify the originating conversation so that the receiving conversation
        knows where to reply to. */
    m.replywith = replywith;

    // Specify the destined conversation name.
    m.inreplyto = inreplyto;
    m.port = connectionPort;

    // To convert object to byte array
    ByteArrayOutputStream bStream = new ByteArrayOutputStream();
    ObjectOutputStream os =
      new ObjectOutputStream(new BufferedOutputStream(bStream));

    this.write(
      "Sending message \"" + m.performative + "\" from " + connectionPort);
    if (m.getContent() != null)
      this.write("Content is: " + m.getContent().getClass().getName());
    os.flush();
    os.writeInt(START_CONVERSATION);
    os.writeObject(m);
    os.flush();

    // Retrives byte array
    byte[] mesg_in_byte = bStream.toByteArray();
    System.out.println("** Multicast Message before encryption in bit:");
    System.out.println("Bit: ");
    for(int i=0; i<mesg_in_byte.length; i++)
    {
      System.out.print(mesg_in_byte[i]);
    }
    // Encrypt information and send
    try
    {
      cipher.init(Cipher.ENCRYPT_MODE, key);
      byte[] encrypted_mesg_in_byte = cipher.doFinal(mesg_in_byte);
      DatagramPacket p =
        new DatagramPacket(
          encrypted_mesg_in_byte,
          encrypted_mesg_in_byte.length,
          InetAddress.getByName(connectionHost),
          connectionPort);
          
      mSocket.send(p);
    }
    catch (InvalidKeyException ike)
    {
      System.out.println("** InvalidKeyException in SecureMulticastHandler");
    }
    catch (BadPaddingException bpe)
    {
      System.out.println("** BadPaddingException in SecureMulticastHandler");
    }
    catch (IllegalBlockSizeException ibse)
    {
      System.out.println(
        "** IllegalBlockSizeException in SecureMulticastHandler");
    }
  }

  /**
   * <p>Method to fetch multicast message from secured multicast message queue
   * destined for this conversation. This method can be considered a blocked 
   * read. This method search message destined for specified conversation name
   * by comparing the String parameter with the inreplyto field in the message.
   * If there is no message for specified conversation, it wait for 1000 
   * miilisecond and try again.
   * @param conversation_name - name of conversation to retrive message
   * @return ksu.cis.mom.Message
   */
  public Message readMessage(String conversation_name)
  {
    while(true)
    {
      for (int i = 0; i < secure_multicast_queue.size(); i++)
      {
        m = (Message) secure_multicast_queue.elementAt(i);
        if (m.inreplyto.equals(conversation_name))
        {
          secure_multicast_queue.removeElementAt(i);
          System.out.println("Received " + m.performative + " message");
          return m;
        }
      }
      try
      {
        Thread.sleep(1000);
      }
      catch(InterruptedException ex)
      {
        this.write("Error in MulticastConversation: " + ex.toString());
      }
    }
  }

  /**
   * <p>Method to fetch multicast message from secured multicast message queue
   * destined for this conversation. This method can be considered a nonblocked 
   * read. This method search message destined for specified conversation name
   * by comparing the String parameter with the inreplyto field in the message.
   * If there is no message for specified conversation, it wait for the amount
   * of specified timeout in miilisecond and then try again. If there is still 
   * no message on the second try, it returns Message with content field "timeout". 
   * @param conversation_name - name of conversation to retrive message
   * @param timeout - read timeout in miilisecond
   * @return Message with content field "timeout" if there is no message 
   * destine for specified conversation.
   */
  public Message nonblockedReadMessage(String conversation_name, long timeout)
  {
    int count = 0;
    while(count < 2)
    {
      for (int i = 0; i < secure_multicast_queue.size(); i++)
      {
        m = (Message) secure_multicast_queue.elementAt(i);
        if (m.inreplyto.equals(conversation_name))
        {
          secure_multicast_queue.removeElementAt(i);
          System.out.println("Received " + m.performative + " message");
          return m;
        }
      }
      try
      {
        Thread.sleep(timeout);
        count++;
      }
      catch(InterruptedException ex)
      {
        this.write("Error in MulticastConversation: " + ex.toString());
      }
    }
    m.setContent("timeout");
    return m;
  }

  /**
   * <p>run method for MulticastConversation class. Each derived conversation 
   * must be run as separate thread and must implement this method regarding to 
   * the type of conversation, initiator or repondent conversation.
   * @see java.lang.Runnable#run()
   */
  public abstract void run();

}// End of SecureMulticastConversation class
