/*
 * @(#)SecureUnicastHandler.java
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */

package ksu.cis.mom;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import javax.net.ssl.SSLServerSocket;
import javax.net.ssl.SSLServerSocketFactory;
import javax.net.ssl.SSLSocket;

/**
 * <p>The SecureUnicastHandler class is used to handle secured unicast 
 * connection from other agents. It uses the Secure Socket Layers 
 * (SSLSocket class) and Secure Socket Layers Server Socket
 * (SSLServerSocket class) to establish a secured connection over TCP/IP stream. 
 * When an agent is created, it can create a new secured unicast handler thread 
 * to be able to receive a start of secured unicast conversation from the 
 * other agents. Below is how an agent can start the SecureUnicastHandler
 * class.
 * <pre>
 * SecureUnicastHandler suh = new SecureUnicastHandler(this.port, this);
 * suh.start();
 * </pre>
 * <p>To create the SecureUnicastHandler, it requires two parameters, port 
 * number and a reference to parent agent object. When it is started, the 
 * secured unicast handler starts a ssl socket server on the indicated port and 
 * waits for a connection from another agent. When a connection is received, 
 * the SecureUnicastHandler simply accepts and calls the parent agents 
 * receiveMessage method with the connection ssl socket and the input and 
 * output streams. The agent then verify the received message and start an 
 * appropriate conversation. 
 * 
 * <p>History:<br>
 * 08/13/2003: created date<br>
 * 09/02/2003: Add more comments suitable for JavaDoc.<br>
 * @author Soctt DeLoach and Chairoj Mekprasertvit
 * @version 1.0, 08/23/2003
 * @since agentMom 1.2.1
 */

public class SecureUnicastHandler extends Thread
{
  /**
   * Port number that the SecureUnicastHandler listens for secured unicast 
   * conversation.
   */
  public int portNo;
  
  /**
   * SSLServerSocket class used for accepting/constructing secured unicast
   * conversation.
   */
  public SSLServerSocket sslServer;
  
  /**
   * Reference to parent agent object that use this SecureUnicastHandler class.
   */
  public Agent parent;
  
  /**
   * Flag control receive thread execution. 
   */
  private boolean alive;

  /**
   * SecureUnicastHandler Constructor
   * @param port - Integer port number
   * @param p - Reference to parent agent object that use this class.
   */
  public SecureUnicastHandler(int port, Agent p)
  {
    // Call to superclass(thread class) constructor
    super("SecureUnicastHandler");
    this.alive = true; 
    parent = (Agent) p;
    if (port <= 1024)
      throw new IllegalArgumentException(
        " ** Bad port number given to Listener constructor for agent "
          + p.name);

    // Try making a SSLServerSocket to the given port
    try
    {
      SSLServerSocketFactory sslSrvFact = 
        (SSLServerSocketFactory) SSLServerSocketFactory.getDefault();
      
      sslServer =(SSLServerSocket)sslSrvFact.createServerSocket(port);
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + p.name
          + " Failed to start SSLServerSocket on port "
          + port
          + e.toString());
      System.exit(1);
    }

    portNo = port;
    System.out.println(
      "Agent " + p.name + " SecureUnicastHandler Started on port " + port);
  }
  
  /**
   * <p>Run method for SecureMulticastHandler.
   * This method listens for a connection starting a new secured unicast 
   * conversation from other agents. Then, initialize input and output streams,
   * and call the parent agent's receiveSecureUnicastConversation method.
   * Finally, the connection, input stream and output stream is passed to the 
   * parent agent's receiveMessage message.
   * @see java.lang.Runnable#run()
   */
  public void run()
  {
    SSLSocket connection;
    int counter = 1;
    ObjectOutputStream output;
    ObjectInputStream input;
    System.out.println("SSLServer Running on " + portNo);

    while (alive)
    {
      try
      {
        connection = (SSLSocket) sslServer.accept();
        System.out.println(
          "Connection "
            + counter++
            + " received from "
            + connection.getInetAddress().getHostName());
        output = new ObjectOutputStream(connection.getOutputStream());
        output.flush();
        input = new ObjectInputStream(connection.getInputStream());
        parent.receiveSecureUnicastConversation(connection, input, output);
      }
      catch (InterruptedIOException ioex)
      {
        System.out.println(" ** Socket Server timeout on port " + portNo);
      }
      catch (IOException cnfex)
      {
        System.out.println(" ** IOException on port " + portNo + cnfex.toString());
      }
    }// end while loop
  }// end run method
  
  /**
   * Set value for isAlive.
   * @param alive - if value is false will make the BroadcastHandler thread to
   * stop the next time this value is checked.
   */
  public void setAlive(boolean alive)
  {
    this.alive = alive;
  }
  
}// end SecureUnicastHandler class